@@ -1,9 +1,29 @@
 /*
- * $Id: fcgi_config.c,v 1.22 2000/04/29 21:24:31 robs Exp $
+ * $Id: fcgi_config.c,v 1.49 2003/10/30 01:08:34 robs Exp $
  */
 
+#define CORE_PRIVATE
 #include "fcgi.h"
 
+#ifdef APACHE2
+
+#include <limits.h>
+#include "mpm_common.h"     /* ap_uname2id, ap_gname2id */
+
+#ifdef WIN32
+#include <direct.h>
+#else
+#include <unistd.h>
+#include "unixd.h"
+#endif
+
+#endif
+
+#ifdef WIN32
+/* warning C4100: unreferenced formal parameter */
+/* warning C4706: assignment within conditional expression */ 
+#pragma warning( disable : 4100 4706 )
+#endif
 
 /*******************************************************************************
  * Get the next configuration directive argument, & return an in_addr and port.
@@ -67,6 +87,30 @@ static const char *get_u_short(pool *p, const char **arg,
     return NULL;
 }
 
+static const char *get_int(pool *p, const char **arg, int *num, int min)
+{
+    char *cp;
+    const char *val = ap_getword_conf(p, arg);
+
+    if (*val == '\0')
+    {
+        return "\"\"";
+    }
+
+    *num = (int) strtol(val, &cp, 10);
+
+    if (*cp != '\0')
+    {
+        return ap_pstrcat(p, "can't parse ", "\"", val, "\"", NULL);
+    }
+    else if (*num < min)
+    {
+        return ap_psprintf(p, "\"%d\" must be >= %d", *num, min);
+    }
+            
+    return NULL;
+}
+
 /*******************************************************************************
  * Get the next configuration directive argument, & return an u_int.
  * The pool arg should be temporary storage.
@@ -180,7 +224,9 @@ static const char *invalid_value(pool *p, const char *cmd, const char *id,
 const char *fcgi_config_set_fcgi_uid_n_gid(int set)
 {
     static int isSet = 0;
+
 #ifndef WIN32
+
     uid_t uid = geteuid();
     gid_t gid = getegid();
 
@@ -191,8 +237,13 @@ const char *fcgi_config_set_fcgi_uid_n_gid(int set)
         return NULL;
     }
 
-    uid = uid ? uid : ap_user_id;
-    gid = uid ? gid : ap_group_id;
+    if (uid == 0) {
+        uid = ap_user_id;
+    }
+
+    if (gid == 0) {
+        gid = ap_group_id;
+    }
 
     if (isSet && (uid != fcgi_user_id || gid != fcgi_group_id)) {
         return "User/Group commands must preceed FastCGI server definitions";
@@ -201,17 +252,19 @@ const char *fcgi_config_set_fcgi_uid_n_gid(int set)
     isSet = 1;
     fcgi_user_id = uid;
     fcgi_group_id = gid;
-#endif
+
+#endif /* !WIN32 */
+
     return NULL;
 }
 
-void fcgi_config_reset_globals(void* dummy)
+apcb_t fcgi_config_reset_globals(void* dummy)
 {
     fcgi_config_pool = NULL;
     fcgi_servers = NULL;
     fcgi_config_set_fcgi_uid_n_gid(0);
-    fcgi_suexec = NULL;
-    fcgi_socket_dir = DEFAULT_SOCK_DIR;
+    fcgi_wrapper = NULL;
+    fcgi_socket_dir = NULL;
     
     fcgi_dynamic_total_proc_count = 0;
     fcgi_dynamic_epoch = 0;
@@ -223,8 +276,8 @@ void fcgi_config_reset_globals(void* dummy)
     dynamicKillInterval = FCGI_DEFAULT_KILL_INTERVAL;
     dynamicUpdateInterval = FCGI_DEFAULT_UPDATE_INTERVAL;
     dynamicGain = FCGI_DEFAULT_GAIN;
-    dynamicThreshhold1 = FCGI_DEFAULT_THRESHHOLD_1;
-    dynamicThreshholdN = FCGI_DEFAULT_THRESHHOLD_N;
+    dynamicThreshold1 = FCGI_DEFAULT_THRESHOLD_1;
+    dynamicThresholdN = FCGI_DEFAULT_THRESHOLD_N;
     dynamicPleaseStartDelay = FCGI_DEFAULT_START_PROCESS_DELAY;
     dynamicAppConnectTimeout = FCGI_DEFAULT_APP_CONN_TIMEOUT;
     dynamicEnvp = &fcgi_empty_env;
@@ -236,6 +289,21 @@ void fcgi_config_reset_globals(void* dummy)
     dynamicRestartDelay = FCGI_DEFAULT_RESTART_DELAY;
     dynamic_pass_headers = NULL;
     dynamic_idle_timeout = FCGI_DEFAULT_IDLE_TIMEOUT;
+	dynamicFlush = FCGI_FLUSH;
+
+#ifndef WIN32
+	/* Close any old pipe (HUP/USR1) */
+	if (fcgi_pm_pipe[0] != -1) {
+		close(fcgi_pm_pipe[0]);
+		fcgi_pm_pipe[0] = -1;
+	}
+	if (fcgi_pm_pipe[1] != -1) {
+		close(fcgi_pm_pipe[1]);
+		fcgi_pm_pipe[1] = -1;
+	}
+#endif
+
+    return APCB_OK;
 }
 
 /*******************************************************************************
@@ -305,11 +373,9 @@ const char *fcgi_config_make_dir(pool *tp, char *path)
  * Create a "dynamic" subdirectory.  If the directory
  * already exists we don't mess with it unless 'wax' is set.
  */
+#ifndef WIN32
 const char *fcgi_config_make_dynamic_dir(pool *p, const int wax)
 {
-#ifndef WIN32
-    DIR *dp = NULL;
-    struct dirent *dirp = NULL;
     const char *err;
     pool *tp;
 
@@ -322,46 +388,79 @@ const char *fcgi_config_make_dynamic_dir(pool *p, const int wax)
     if (!wax)
         return NULL;
 
-    /* Create a subpool for the directory operations */
-    tp = ap_make_sub_pool(p);
+#ifdef APACHE2
+    {
+        apr_dir_t * dir;
+        apr_finfo_t finfo;
 
-    dp = ap_popendir(tp, fcgi_dynamic_dir);
-    if (dp == NULL) {
-        ap_destroy_pool(tp);
-        return ap_psprintf(p, "can't open dynamic directory \"%s\": %s",
-            fcgi_dynamic_dir, strerror(errno));
+        if (apr_pool_create(&tp, p))
+            return "apr_pool_create() failed";
+
+        if (apr_dir_open(&dir, fcgi_dynamic_dir, tp))
+            return "apr_dir_open() failed";
+
+        /* delete the contents */
+
+        while (apr_dir_read(&finfo, APR_FINFO_NAME, dir) == APR_SUCCESS)
+        {
+            if (strcmp(finfo.name, ".") == 0 || strcmp(finfo.name, "..") == 0)
+                continue;
+
+            apr_file_remove(finfo.name, tp);
+        }
     }
 
-    /* Delete everything in the directory, its all FCGI specific */
-    while ((dirp = readdir(dp)) != NULL) {
-        if (strcmp(dirp->d_name, ".") == 0
-                || strcmp(dirp->d_name, "..") == 0) {
-            continue;
+#else /* !APACHE2 */
+    {
+        DIR *dp;
+        struct dirent *dirp = NULL;
+
+        tp = ap_make_sub_pool(p);
+
+        dp = ap_popendir(tp, fcgi_dynamic_dir);
+        if (dp == NULL) {
+            ap_destroy_pool(tp);
+            return ap_psprintf(p, "can't open dynamic directory \"%s\": %s",
+                fcgi_dynamic_dir, strerror(errno));
         }
 
-        unlink(ap_pstrcat(tp, fcgi_dynamic_dir, "/", dirp->d_name, NULL));
+        /* delete the contents */
+
+        while ((dirp = readdir(dp)) != NULL) 
+        {
+            if (strcmp(dirp->d_name, ".") == 0 || strcmp(dirp->d_name, "..") == 0)
+                continue;
+
+            unlink(ap_pstrcat(tp, fcgi_dynamic_dir, "/", dirp->d_name, NULL));
+        }
     }
 
+#endif /* !APACHE2 */
+
     ap_destroy_pool(tp);
 
-#else
-    fcgi_dynamic_dir = ap_pstrcat(p, fcgi_socket_dir, "dynamic", NULL);
-#endif
     return NULL;
 }
-
+#endif
 
 /*******************************************************************************
  * Change the directory used for the Unix/Domain sockets from the default.
  * Create the directory and the "dynamic" subdirectory.
  */
-const char *fcgi_config_set_socket_dir(cmd_parms *cmd, void *dummy, char *arg)
+const char *fcgi_config_set_socket_dir(cmd_parms *cmd, void *dummy, const char *arg)
 {
     pool * const tp = cmd->temp_pool;
     const char * const name = cmd->cmd->name;
     const char *err;
+    char * arg_nc;
 
-    if (strcmp(fcgi_socket_dir, DEFAULT_SOCK_DIR) != 0) {
+    err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err)
+    {
+        return err;
+    }
+
+    if (fcgi_socket_dir) {
         return ap_psprintf(tp, "%s %s: already defined as \"%s\"",
                         name, arg, fcgi_socket_dir);
     }
@@ -376,44 +475,69 @@ const char *fcgi_config_set_socket_dir(cmd_parms *cmd, void *dummy, char *arg)
             name);
     }
 
+    arg_nc = ap_pstrdup(cmd->pool, arg);
+
 #ifndef WIN32
-    arg = ap_os_canonical_filename(cmd->pool, arg);
-    arg = ap_server_root_relative(cmd->pool, arg);
+
+#ifdef APACHE2
+    if (apr_filepath_merge(&arg_nc, "", arg, 0, cmd->pool))
+        return ap_psprintf(tp, "%s %s: invalid filepath", name, arg);
 #else
-	if (strncmp(arg, "\\\\.\\pipe\\", 9) != 0)
-		return ap_psprintf(tp, "%s %s is invalid format",name, arg);
+    arg_nc = ap_os_canonical_filename(cmd->pool, arg_nc);
 #endif
 
-    fcgi_socket_dir = arg;
+    arg_nc = ap_server_root_relative(cmd->pool, arg_nc);
 
-#ifndef WIN32
+#else /* WIN32 */
+
+	if (strncmp(arg_nc, "\\\\.\\pipe\\", 9) != 0)
+		return ap_psprintf(tp, "%s %s is invalid format",name, arg_nc);
+
+#endif
+
+    fcgi_socket_dir = arg_nc;
+
+#ifdef WIN32
+    fcgi_dynamic_dir = ap_pstrcat(cmd->pool, fcgi_socket_dir, "dynamic", NULL);
+#else
     err = fcgi_config_make_dir(tp, fcgi_socket_dir);
     if (err != NULL)
-        return ap_psprintf(tp, "%s %s: %s", name, arg, err);
+        return ap_psprintf(tp, "%s %s: %s", name, arg_nc, err);
 
     err = fcgi_config_make_dynamic_dir(cmd->pool, 0);
     if (err != NULL)
-        return ap_psprintf(tp, "%s %s: %s", name, arg, err);
+        return ap_psprintf(tp, "%s %s: %s", name, arg_nc, err);
 #endif
 
     return NULL;
 }
 
 /*******************************************************************************
- * Enable, disable, or specify the path to the suexec program.
+ * Enable, disable, or specify the path to a wrapper used to invoke all
+ * FastCGI applications.
  */
-const char *fcgi_config_set_suexec(cmd_parms *cmd, void *dummy, const char *arg)
+const char *fcgi_config_set_wrapper(cmd_parms *cmd, void *dummy, const char *arg)
 {
+#ifdef WIN32
+    return ap_psprintf(cmd->temp_pool, 
+        "the %s directive is not supported on WIN", cmd->cmd->name);
+#else
+
     const char *err = NULL;
     const char * const name = cmd->cmd->name;
     pool * const tp = cmd->temp_pool;
-    char * suexec;
+    char * wrapper = NULL;
 
-    if (!ap_suexec_enabled) {
-        if (strcasecmp(arg, "Off") != 0) {
-	        fprintf(stderr, "Warning: %s requires SUEXEC wrapper be enabled in Apache\n", name);
-	    }
-	    return NULL;
+    err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err)
+    {
+        return err;
+    }
+
+    if (fcgi_wrapper)
+    {
+        return ap_psprintf(tp, "%s was already set to \"%s\"",
+                           name, fcgi_wrapper);
     }
 
     err = fcgi_config_set_fcgi_uid_n_gid(1);
@@ -425,31 +549,39 @@ const char *fcgi_config_set_suexec(cmd_parms *cmd, void *dummy, const char *arg)
             "The %s command must preceed static FastCGI server definitions", name);
     }
 
-    if (strcasecmp(arg, "On") == 0) {
-        fcgi_suexec = SUEXEC_BIN;
+    if (strcasecmp(arg, "Off") == 0) {
+        fcgi_wrapper = NULL;
+        return NULL;
     }
-    else if (strcasecmp(arg, "Off") == 0) {
-        fcgi_suexec = NULL;
-    }
-    else {
-        suexec = (char *) ap_os_canonical_filename(cmd->pool, arg);
-        suexec = ap_server_root_relative(cmd->pool, suexec);
 
-#ifdef WIN32
-        err = fcgi_util_check_access(tp, suexec, NULL, _S_IEXEC, fcgi_user_id, fcgi_group_id);
+    if (strcasecmp(arg, "On") == 0) 
+    {
+        wrapper = SUEXEC_BIN;
+    }
+    else
+    {
+#ifdef APACHE2
+        if (apr_filepath_merge(&wrapper, "", arg, 0, cmd->pool))
+            return ap_psprintf(tp, "%s %s: invalid filepath", name, arg);
 #else
-        err = fcgi_util_check_access(tp, suexec, NULL, X_OK, fcgi_user_id, fcgi_group_id);
+        wrapper = ap_os_canonical_filename(cmd->pool, (char *) arg);
 #endif
 
-        if (err != NULL) {
-            return ap_psprintf(tp,
-                "%s: \"%s\" access for server (uid %ld, gid %ld) failed: %s",
-                name, suexec, (long)fcgi_user_id, (long)fcgi_group_id, err);
-        }
-
-        fcgi_suexec = suexec;
+        wrapper = ap_server_root_relative(cmd->pool, wrapper);
     }
+
+    err = fcgi_util_check_access(tp, wrapper, NULL, X_OK, fcgi_user_id, fcgi_group_id);
+    if (err) 
+    {
+        return ap_psprintf(tp, "%s: \"%s\" execute access for server "
+                           "(uid %ld, gid %ld) failed: %s", name, wrapper,
+                           (long) fcgi_user_id, (long) fcgi_group_id, err);
+    }
+
+    fcgi_wrapper = wrapper;
+
     return NULL;
+#endif /* !WIN32 */
 }
 
 /*******************************************************************************
@@ -467,27 +599,43 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
     char **envp = ap_pcalloc(tp, sizeof(char *) * (MAX_INIT_ENV_VARS + 3));
     unsigned int envc = 0;
 
+#ifdef WIN32
+    HANDLE mutex;
+#endif
+
+    err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err)
+    {
+        return err;
+    }
+
     if (*fs_path == '\0')
         return "AppClass requires a pathname!?";
 
     if ((err = fcgi_config_set_fcgi_uid_n_gid(1)) != NULL)
         return ap_psprintf(tp, "%s %s: %s", name, fs_path, err);
-        
+
+#ifdef APACHE2
+    if (apr_filepath_merge(&fs_path, "", fs_path, 0, p))
+        return ap_psprintf(tp, "%s %s: invalid filepath", name, fs_path);
+#else
     fs_path = ap_os_canonical_filename(p, fs_path);
+#endif
     fs_path = ap_server_root_relative(p, fs_path);
 
     ap_getparents(fs_path);
     ap_no2slash(fs_path);
 
     /* See if we've already got one of these configured */
-    s = fcgi_util_fs_get_by_id(fs_path, cmd->server->server_uid,
-                       cmd->server->server_gid);
+    s = fcgi_util_fs_get_by_id(fs_path, fcgi_util_get_server_uid(cmd->server),
+                               fcgi_util_get_server_gid(cmd->server));
     if (s != NULL) {
-        if (fcgi_suexec) {
+        if (fcgi_wrapper) {
             return ap_psprintf(tp,
-                "%s: redefinition of a previously defined FastCGI server \"%s\" with uid=%ld and gid=%ld",
-                name, fs_path, (long)cmd->server->server_uid,
-                (long)cmd->server->server_gid);
+                "%s: redefinition of a previously defined FastCGI "
+                "server \"%s\" with uid=%ld and gid=%ld",
+                name, fs_path, (long) fcgi_util_get_server_uid(cmd->server),
+                (long) fcgi_util_get_server_gid(cmd->server));
         }
         else {
             return ap_psprintf(tp,
@@ -496,8 +644,7 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
         }
     }
 
-    err = fcgi_util_fs_is_path_ok(tp, fs_path, NULL, cmd->server->server_uid,
-                          cmd->server->server_gid);
+    err = fcgi_util_fs_is_path_ok(tp, fs_path, NULL);
     if (err != NULL) {
         return ap_psprintf(tp, "%s: \"%s\" %s", name, fs_path, err);
     }
@@ -509,35 +656,24 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
     s->numProcesses = 1;
 
 #ifdef WIN32
-    // TCP FastCGI applications require SystemRoot be present in the environment
-    // Put it in both for consistency to the application
-    fcgi_config_set_env_var(tp, envp, &envc, "SystemRoot");
-#else
-    if (fcgi_suexec) {
-        struct passwd *pw;
-        struct group  *gr;
 
-        s->uid = cmd->server->server_uid;
-        pw = getpwuid(s->uid);
-        if (pw == NULL) {
-            return ap_psprintf(tp, "mod_fastcgi: "
-                "getpwuid() couldn't determine the username for uid '%ld', "
-                "you probably need to modify the User directive: %s",
-                (long)s->uid, strerror(errno));
-        }
-        s->user = ap_pstrdup(p, pw->pw_name);
-        s->username = s->user;
+    /* TCP FastCGI applications require SystemRoot be present in the environment
+     * Put it in both for consistency to the application */
+    fcgi_config_set_env_var(p, envp, &envc, "SystemRoot");
 
-        s->gid = cmd->server->server_gid;
-        gr = getgrgid(s->gid);
-        if (gr == NULL) {
-            return ap_psprintf(tp, "mod_fastcgi: "
-                "getgrgid() couldn't determine the group name for gid '%ld', "
-                "you probably need to modify the Group directive: %s\n",
-                (long)s->gid, strerror(errno));
-        }
-        s->group = ap_pstrdup(p, gr->gr_name);
+    mutex = CreateMutex(NULL, FALSE, fs_path);
+    
+    if (mutex == NULL)
+    {
+        ap_log_error(FCGI_LOG_ALERT, fcgi_apache_main_server,
+            "FastCGI: CreateMutex() failed");
+        return "failed to create FastCGI application accept mutex";
     }
+    
+    SetHandleInformation(mutex, HANDLE_FLAG_INHERIT, TRUE);
+
+    s->mutex_env_string = ap_psprintf(p, "_FCGI_MUTEX_=%ld", mutex);
+
 #endif
 
     /*  Parse directive arguments */
@@ -553,7 +689,7 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
                 return invalid_value(tp, name, fs_path, option, err);
         }
         else if (strcasecmp(option, "-init-start-delay") == 0) {
-            if ((err = get_u_int(tp, &arg, &s->initStartDelay, 0)))
+            if ((err = get_int(tp, &arg, &s->initStartDelay, 0)))
                 return invalid_value(tp, name, fs_path, option, err);
         }
         else if (strcasecmp(option, "-priority") == 0) {
@@ -592,20 +728,68 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
         else if (strcasecmp(option, "-flush") == 0) {
             s->flush = 1;
         }
+        else if (strcasecmp(option, "-user") == 0) {
+#ifdef WIN32
+            return ap_psprintf(tp, 
+                "%s %s: the -user option isn't supported on WIN", name, fs_path);
+#else
+            s->user = ap_getword_conf(tp, &arg);
+            if (*s->user == '\0')
+                return invalid_value(tp, name, fs_path, option, "\"\"");
+#endif
+        }
+        else if (strcasecmp(option, "-group") == 0) {
+#ifdef WIN32
+            return ap_psprintf(tp, 
+                "%s %s: the -group option isn't supported on WIN", name, fs_path);
+#else
+            s->group = ap_getword_conf(tp, &arg);
+            if (*s->group == '\0')
+                return invalid_value(tp, name, fs_path, option, "\"\"");
+#endif
+        }
         else {
             return ap_psprintf(tp, "%s %s: invalid option: %s", name, fs_path, option);
         }
     } /* while */
 
+#ifndef WIN32
+    if (fcgi_wrapper)
+    {
+        if (s->group == NULL)
+        {
+            s->group = ap_psprintf(tp, "#%ld", fcgi_util_get_server_gid(cmd->server));
+        }
+
+        if (s->user == NULL)
+        {
+            s->user = ap_psprintf(p, "#%ld", fcgi_util_get_server_uid(cmd->server)); 
+        }
+
+        s->uid = ap_uname2id(s->user);
+        s->gid = ap_gname2id(s->group);
+    }
+    else if (s->user || s->group)
+    {
+        ap_log_error(FCGI_LOG_WARN, cmd->server, "FastCGI: there is no "
+                     "fastcgi wrapper set, user/group options are ignored");
+    }
+
+    if ((err = fcgi_util_fs_set_uid_n_gid(p, s, s->uid, s->gid)))
+    {
+        return ap_psprintf(tp, 
+            "%s %s: invalid user or group: %s", name, fs_path, err);
+    }
+#endif /* !WIN32 */
+
     if (s->socket_path != NULL && s->port != 0) {
         return ap_psprintf(tp,
                 "%s %s: -port and -socket are mutually exclusive options",
                 name, fs_path);
     }
 
-    /* Move env array to a surviving pool, leave an extra slot for WIN32 _FCGI_MUTEX_ */
-    ++envc;
-    s->envp = (char **)ap_palloc(p, sizeof(char *) * ++envc);
+    /* Move env array to a surviving pool */
+    s->envp = (char **)ap_pcalloc(p, sizeof(char *) * (envc + 4));
     memcpy(s->envp, envp, sizeof(char *) * envc);
 
     /* Initialize process structs */
@@ -626,6 +810,16 @@ const char *fcgi_config_new_static_server(cmd_parms *cmd, void *dummy, const cha
     } else {
         if (s->socket_path == NULL)
              s->socket_path = fcgi_util_socket_hash_filename(tp, fs_path, s->user, s->group);
+
+        if (fcgi_socket_dir == NULL)
+        {
+#ifdef WIN32
+            fcgi_socket_dir = DEFAULT_SOCK_DIR;
+#else
+            fcgi_socket_dir = ap_server_root_relative(p, DEFAULT_SOCK_DIR);
+#endif
+        }
+
         s->socket_path = fcgi_util_socket_make_path_absolute(p, s->socket_path, 0);
 #ifndef WIN32
         err = fcgi_util_socket_make_domain_addr(p, (struct sockaddr_un **)&s->socket_addr,
@@ -652,27 +846,41 @@ const char *fcgi_config_new_external_server(cmd_parms *cmd, void *dummy, const c
     char *fs_path = ap_getword_conf(p, &arg);
     const char *option, *err;
 
+    err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err)
+    {
+        return err;
+    }
+
     if (!*fs_path) {
         return ap_pstrcat(tp, name, " requires a path and either a -socket or -host option", NULL);
     }
 
+#ifdef APACHE2
+    if (apr_filepath_merge(&fs_path, "", fs_path, 0, p))
+        return ap_psprintf(tp, "%s %s: invalid filepath", name, fs_path);
+#else
     fs_path = ap_os_canonical_filename(p, fs_path);
+#endif
+
     fs_path = ap_server_root_relative(p, fs_path);
 
     ap_getparents(fs_path);
     ap_no2slash(fs_path);
 
     /* See if we've already got one of these bettys configured */
-    s = fcgi_util_fs_get_by_id(fs_path, cmd->server->server_uid,
-                       cmd->server->server_gid);
+    s = fcgi_util_fs_get_by_id(fs_path, fcgi_util_get_server_uid(cmd->server),
+                               fcgi_util_get_server_gid(cmd->server));
     if (s != NULL) {
-        if (fcgi_suexec != NULL) {
+        if (fcgi_wrapper) {
             return ap_psprintf(tp,
-                "%s: redefinition of a previously defined class \"%s\" with uid=%ld and gid=%ld",
-                name, fs_path, (long)cmd->server->server_uid,
-                (long)cmd->server->server_gid);
+                "%s: redefinition of a previously defined class \"%s\" "
+                "with uid=%ld and gid=%ld",
+                name, fs_path, (long) fcgi_util_get_server_uid(cmd->server),
+                (long) fcgi_util_get_server_gid(cmd->server));
         }
-        else {
+        else 
+        {
             return ap_psprintf(tp,
                 "%s: redefinition of previously defined class \"%s\"", name, fs_path);
         }
@@ -682,10 +890,6 @@ const char *fcgi_config_new_external_server(cmd_parms *cmd, void *dummy, const c
     s->fs_path = fs_path;
     s->directive = APP_CLASS_EXTERNAL;
 
-    err = fcgi_util_fs_set_uid_n_gid(p, s, cmd->server->server_uid, cmd->server->server_gid);
-    if (err != NULL)
-        return ap_psprintf(tp, "%s %s: %s", name, fs_path, err);
-
     /*  Parse directive arguments */
     while (*arg != '\0') {
         option = ap_getword_conf(tp, &arg);
@@ -714,11 +918,61 @@ const char *fcgi_config_new_external_server(cmd_parms *cmd, void *dummy, const c
         else if (strcasecmp(option, "-flush") == 0) {
             s->flush = 1;
         }
+        else if (strcasecmp(option, "-user") == 0) {
+#ifdef WIN32
+            return ap_psprintf(tp, 
+                "%s %s: the -user option isn't supported on WIN", name, fs_path);
+#else
+            s->user = ap_getword_conf(tp, &arg);
+            if (*s->user == '\0')
+                return invalid_value(tp, name, fs_path, option, "\"\"");
+#endif
+        }
+        else if (strcasecmp(option, "-group") == 0) {
+#ifdef WIN32
+            return ap_psprintf(tp, 
+                "%s %s: the -group option isn't supported on WIN", name, fs_path);
+#else
+            s->group = ap_getword_conf(tp, &arg);
+            if (*s->group == '\0')
+                return invalid_value(tp, name, fs_path, option, "\"\"");
+#endif
+        }
         else {
             return ap_psprintf(tp, "%s %s: invalid option: %s", name, fs_path, option);
         }
     } /* while */
 
+
+#ifndef WIN32
+    if (fcgi_wrapper)
+    {
+        if (s->group == NULL)
+        {
+            s->group = ap_psprintf(tp, "#%ld", fcgi_util_get_server_gid(cmd->server));
+        }
+
+        if (s->user == NULL)
+        {
+            s->user = ap_psprintf(p, "#%ld", fcgi_util_get_server_uid(cmd->server));
+        }
+
+        s->uid = ap_uname2id(s->user);
+        s->gid = ap_gname2id(s->group);
+    }
+    else if (s->user || s->group)
+    {
+        ap_log_error(FCGI_LOG_WARN, cmd->server, "FastCGI: there is no "
+                     "fastcgi wrapper set, user/group options are ignored");
+    }
+
+    if ((err = fcgi_util_fs_set_uid_n_gid(p, s, s->uid, s->gid)))
+    {
+        return ap_psprintf(tp,
+            "%s %s: invalid user or group: %s", name, fs_path, err);
+    }
+#endif /* !WIN32 */
+
     /* Require one of -socket or -host, but not both */
     if (s->socket_path != NULL && s->port != 0) {
         return ap_psprintf(tp,
@@ -737,6 +991,16 @@ const char *fcgi_config_new_external_server(cmd_parms *cmd, void *dummy, const c
         if (err != NULL)
             return ap_psprintf(tp, "%s %s: %s", name, fs_path, err);
     } else {
+
+        if (fcgi_socket_dir == NULL)
+        {
+#ifdef WIN32
+            fcgi_socket_dir = DEFAULT_SOCK_DIR;
+#else
+            fcgi_socket_dir = ap_server_root_relative(p, DEFAULT_SOCK_DIR);
+#endif
+        }
+
         s->socket_path = fcgi_util_socket_make_path_absolute(p, s->socket_path, 0);
 #ifndef WIN32
         err = fcgi_util_socket_make_domain_addr(p, (struct sockaddr_un **)&s->socket_addr,
@@ -775,6 +1039,12 @@ const char *fcgi_config_set_config(cmd_parms *cmd, void *dummy, const char *arg)
     unsigned int envc = 0;
     char **envp = (char **)ap_pcalloc(tp, sizeof(char *) * (MAX_INIT_ENV_VARS + 3));
 
+    err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err)
+    {
+        return err;
+    }
+
     /* Parse the directive arguments */
     while (*arg) {
         option = ap_getword_conf(tp, &arg);
@@ -784,11 +1054,11 @@ const char *fcgi_config_set_config(cmd_parms *cmd, void *dummy, const char *arg)
                 return invalid_value(tp, name, NULL, option, err);
         }
         else if (strcasecmp(option, "-minProcesses") == 0) {
-            if ((err = get_u_int(tp, &arg, &dynamicMinProcs, 0)))
+            if ((err = get_int(tp, &arg, &dynamicMinProcs, 0)))
                 return invalid_value(tp, name, NULL, option, err);
         }
         else if (strcasecmp(option, "-maxClassProcesses") == 0) {
-            if ((err = get_u_int(tp, &arg, &dynamicMaxClassProcs, 1)))
+            if ((err = get_int(tp, &arg, &dynamicMaxClassProcs, 1)))
                 return invalid_value(tp, name, NULL, option, err);
         }
         else if (strcasecmp(option, "-killInterval") == 0) {
@@ -803,12 +1073,16 @@ const char *fcgi_config_set_config(cmd_parms *cmd, void *dummy, const char *arg)
             if ((err = get_float(tp, &arg, &dynamicGain, 0.0, 1.0)))
                 return invalid_value(tp, name, NULL, option, err);
         }
-        else if (strcasecmp(option, "-singleThreshhold") == 0) {
-            if ((err = get_u_int(tp, &arg, &dynamicThreshhold1, 0)))
+        else if ((strcasecmp(option, "-singleThreshold") == 0)
+		    || (strcasecmp(option, "-singleThreshhold") == 0)) 
+        {
+            if ((err = get_int(tp, &arg, &dynamicThreshold1, 0)))
                 return invalid_value(tp, name, NULL, option, err);
         }
-        else if (strcasecmp(option, "-multiThreshhold") == 0) {
-            if ((err = get_u_int(tp, &arg, &dynamicThreshholdN, 0)))
+        else if ((strcasecmp(option, "-multiThreshold") == 0)
+		    || (strcasecmp(option, "-multiThreshhold") == 0)) 
+        {
+            if ((err = get_int(tp, &arg, &dynamicThresholdN, 0)))
                 return invalid_value(tp, name, NULL, option, err);
         }
         else if (strcasecmp(option, "-startDelay") == 0) {
@@ -852,15 +1126,25 @@ const char *fcgi_config_set_config(cmd_parms *cmd, void *dummy, const char *arg)
         }
         else if (strcasecmp(option, "-autoUpdate") == 0) {
             dynamicAutoUpdate = 1;
+		}
+        else if (strcasecmp(option, "-flush") == 0) {
+            dynamicFlush = TRUE;
         }
         else {
             return ap_psprintf(tp, "%s: invalid option: %s", name, option);
         }
     } /* while */
 
-    /* Move env array to a surviving pool, leave an extra slot for WIN32 _FCGI_MUTEX_ */
-    ++envc;
-    dynamicEnvp = (char **)ap_palloc(p, sizeof(char *) * ++envc);
+    if (dynamicProcessSlack >= dynamicMaxProcs + 1) {
+	    /* the kill policy would work unexpectedly */
+    	return ap_psprintf(tp, 
+            "%s: processSlack (%u) must be less than maxProcesses (%u) + 1", 
+        	name, dynamicProcessSlack, dynamicMaxProcs);
+    }
+
+    /* Move env array to a surviving pool, leave 2 extra slots for 
+     * WIN32 _FCGI_MUTEX_ and _FCGI_SHUTDOWN_EVENT_ */
+    dynamicEnvp = (char **)ap_pcalloc(p, sizeof(char *) * (envc + 4));
     memcpy(dynamicEnvp, envp, sizeof(char *) * envc);
 
     return NULL;
@@ -878,20 +1162,27 @@ void *fcgi_config_create_dir_config(pool *p, char *dummy)
 }
 
 
-const char *fcgi_config_new_auth_server(cmd_parms * const cmd,
-    fcgi_dir_config *dir_config, const char *fs_path, const char * const compat)
+const char *fcgi_config_new_auth_server(cmd_parms * cmd,
+    void * dircfg, const char *fs_path, const char * compat)
 {
+    fcgi_dir_config * dir_config = (fcgi_dir_config *) dircfg;
     pool * const tp = cmd->temp_pool;
-    const uid_t uid = cmd->server->server_uid;
-    const gid_t gid = cmd->server->server_gid;
     char * auth_server;
 
+#ifdef APACHE2
+    if (apr_filepath_merge(&auth_server, "", fs_path, 0, cmd->pool))
+        return ap_psprintf(tp, "%s %s: invalid filepath", cmd->cmd->name, fs_path);
+#else
     auth_server = (char *) ap_os_canonical_filename(cmd->pool, fs_path);
+#endif
+
     auth_server = ap_server_root_relative(cmd->pool, auth_server);
 
     /* Make sure its already configured or at least a candidate for dynamic */
-    if (fcgi_util_fs_get_by_id(auth_server, uid, gid) == NULL) {
-        const char *err = fcgi_util_fs_is_path_ok(tp, auth_server, NULL, uid, gid);
+    if (fcgi_util_fs_get_by_id(auth_server, fcgi_util_get_server_uid(cmd->server),
+                               fcgi_util_get_server_gid(cmd->server)) == NULL) 
+    {
+        const char *err = fcgi_util_fs_is_path_ok(tp, auth_server, NULL);
         if (err)
             return ap_psprintf(tp, "%s: \"%s\" %s", cmd->cmd->name, auth_server, err);
     }
@@ -917,15 +1208,15 @@ const char *fcgi_config_new_auth_server(cmd_parms * const cmd,
     return NULL;
 }
 
-const char *fcgi_config_set_authoritative_slot(const cmd_parms * const cmd,
-    fcgi_dir_config * const dir_config, int arg)
+const char *fcgi_config_set_authoritative_slot(cmd_parms * cmd,
+    void * dir_config, int arg)
 {
     int offset = (int)(long)cmd->info;
 
     if (arg)
-        *(int *)(dir_config + offset) |= FCGI_AUTHORITATIVE;
+        *((u_char *)dir_config + offset) |= FCGI_AUTHORITATIVE;
     else
-        *(int *)(dir_config + offset) &= ~FCGI_AUTHORITATIVE;
+        *((u_char *)dir_config + offset) &= ~FCGI_AUTHORITATIVE;
 
     return NULL;
 }
