@@ -1,11 +1,18 @@
 /*
- * $Id: fcgi_protocol.c,v 1.17 2000/04/27 19:53:54 robs Exp $
+ * $Id: fcgi_protocol.c,v 1.25 2003/02/03 22:59:01 robs Exp $
  */
 
-
 #include "fcgi.h"
 #include "fcgi_protocol.h"
 
+#ifdef APACHE2
+#include "apr_lib.h"
+#endif
+
+#ifdef WIN32
+#pragma warning( disable : 4706)
+#endif
+
  /*******************************************************************************
  * Build and queue a FastCGI message header.  It is the caller's
  * responsibility to make sure that there's enough space in the buffer, and
@@ -16,9 +23,10 @@ static void queue_header(fcgi_request *fr, unsigned char type, unsigned int len)
 {
     FCGI_Header header;
 
-    ap_assert(type > 0 && type <= FCGI_MAXTYPE);
-    ap_assert(len >= 0 && len <= 0xffff);
-    ap_assert(BufferFree(fr->serverOutputBuffer) > sizeof(FCGI_Header));
+    ASSERT(type > 0);
+    ASSERT(type <= FCGI_MAXTYPE);
+    ASSERT(len <= 0xffff);
+    ASSERT(BufferFree(fr->serverOutputBuffer) >= sizeof(FCGI_Header));
 
     /* Assemble and queue the packet header. */
     header.version = FCGI_VERSION;
@@ -38,7 +46,7 @@ static void queue_header(fcgi_request *fr, unsigned char type, unsigned int len)
 static void build_begin_request(unsigned int role, unsigned char keepConnection,
         FCGI_BeginRequestBody *body)
 {
-    ap_assert((role >> 16) == 0);
+    ASSERT((role >> 16) == 0);
     body->roleB1 = (unsigned char) (role >>  8);
     body->roleB0 = (unsigned char) role;
     body->flags = (unsigned char) ((keepConnection) ? FCGI_KEEP_CONN : 0);
@@ -54,7 +62,7 @@ void fcgi_protocol_queue_begin_request(fcgi_request *fr)
     int bodySize = sizeof(FCGI_BeginRequestBody);
 
     /* We should be the first ones to use this buffer */
-    ap_assert(BufferLength(fr->serverOutputBuffer) == 0);
+    ASSERT(BufferLength(fr->serverOutputBuffer) == 0);
 
     build_begin_request(fr->role, FALSE, &body);
     queue_header(fr, FCGI_BEGIN_REQUEST, bodySize);
@@ -69,7 +77,7 @@ static void build_env_header(int nameLen, int valueLen,
 {
     unsigned char *startHeaderBuffPtr = headerBuffPtr;
 
-    ap_assert(nameLen >= 0);
+    ASSERT(nameLen >= 0);
 
     if (nameLen < 0x80) {
         *headerBuffPtr++ = (unsigned char) nameLen;
@@ -80,7 +88,7 @@ static void build_env_header(int nameLen, int valueLen,
         *headerBuffPtr++ = (unsigned char) nameLen;
     }
 
-    ap_assert(valueLen >= 0);
+    ASSERT(valueLen >= 0);
 
     if (valueLen < 0x80) {
         *headerBuffPtr++ = (unsigned char) valueLen;
@@ -205,7 +213,7 @@ int fcgi_protocol_queue_env(request_rec *r, fcgi_request *fr, env_status *env)
         {
         case PREP:
             env->equalPtr = strchr(*env->envp, '=');
-            ap_assert(env->equalPtr != NULL);
+            ASSERT(env->equalPtr != NULL);
             env->nameLen = env->equalPtr - *env->envp;
             env->valueLen = strlen(++env->equalPtr);
             build_env_header(env->nameLen, env->valueLen, env->headerBuff, &env->headerLen);
@@ -316,13 +324,13 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                 ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r,
                     "FastCGI: comm with server \"%s\" aborted: protocol error: invalid version: %d != FCGI_VERSION(%d)",
                     fr->fs_path, header.version, FCGI_VERSION);
-                return SERVER_ERROR;
+                return HTTP_INTERNAL_SERVER_ERROR;
             }
             if (header.type > FCGI_MAXTYPE) {
                 ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r,
                     "FastCGI: comm with server \"%s\" aborted: protocol error: invalid type: %d > FCGI_MAXTYPE(%d)",
                     fr->fs_path, header.type, FCGI_MAXTYPE);
-                return SERVER_ERROR;
+                return HTTP_INTERNAL_SERVER_ERROR;
             }
 
             fr->packetType = header.type;
@@ -336,7 +344,7 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
          * State #2:  got a header, and processing packet bytes.
          */
         len = min(fr->dataLen, BufferLength(fr->serverInputBuffer));
-        ap_assert(len >= 0);
+        ASSERT(len >= 0);
         switch (fr->packetType) {
             case FCGI_STDOUT:
                 if (len > 0) {
@@ -354,6 +362,7 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                             break;
                         default:
                             /* Toss data on the floor */
+                            fcgi_buf_removed(fr->serverInputBuffer, len);
                             break;
                     }
                     fr->dataLen -= len;
@@ -397,10 +406,12 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                     {
                         if (start != end)
                         {
-                        	*end = '\0';
-                        	ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r, "FastCGI: server \"%s\" stderr: %s", fr->fs_path, start);
-                        } 
-                        end += strspn(++end, "\r\n");
+                            *end = '\0';
+                            ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r, 
+                                "FastCGI: server \"%s\" stderr: %s", fr->fs_path, start);
+                        }
+                        end++;
+                        end += strspn(end, "\r\n");
                         fr->fs_stderr_len -= (end - start);
                         start = end;
                     }
@@ -415,7 +426,8 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                         else if (fr->fs_stderr_len == FCGI_SERVER_MAX_STDERR_LINE_LEN)
                         {
                             /* Full buffer, dump it and complain */
-                            ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r, "FastCGI: server \"%s\" stderr: %s", fr->fs_path, fr->fs_stderr);
+                            ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r, 
+                               "FastCGI: server \"%s\" stderr: %s", fr->fs_path, fr->fs_stderr);
                             ap_log_rerror(FCGI_LOG_WARN_NOERRNO, fr->r,
                                 "FastCGI: too much stderr received from server \"%s\", "
                                 "increase FCGI_SERVER_MAX_STDERR_LINE_LEN (%d) and rebuild "
@@ -431,10 +443,11 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                 if (!fr->readingEndRequestBody) {
                     if (fr->dataLen != sizeof(FCGI_EndRequestBody)) {
                         ap_log_rerror(FCGI_LOG_ERR_NOERRNO, fr->r,
-                            "FastCGI: comm with server \"%s\" aborted: protocol error: invalid FCGI_END_REQUEST size: "
+                            "FastCGI: comm with server \"%s\" aborted: protocol error: "
+                            "invalid FCGI_END_REQUEST size: "
                             "%d != sizeof(FCGI_EndRequestBody)(%d)",
                             fr->fs_path, fr->dataLen, sizeof(FCGI_EndRequestBody));
-                        return SERVER_ERROR;
+                        return HTTP_INTERNAL_SERVER_ERROR;
                     }
                     fr->readingEndRequestBody = TRUE;
                 }
@@ -455,7 +468,7 @@ int fcgi_protocol_dequeue(pool *p, fcgi_request *fr)
                             "FastCGI: comm with server \"%s\" aborted: protocol error: invalid FCGI_END_REQUEST status: "
                             "%d != FCGI_REQUEST_COMPLETE(%d)", fr->fs_path,
                             erBody->protocolStatus, FCGI_REQUEST_COMPLETE);
-                        return SERVER_ERROR;
+                        return HTTP_INTERNAL_SERVER_ERROR;
                     }
                     fr->exitStatus = (erBody->appStatusB3 << 24)
                         + (erBody->appStatusB2 << 16)
